@@ -1,3 +1,4 @@
+import inspect
 from abc import ABC
 import numpy as np
 from typing import Optional, Union, Tuple
@@ -43,6 +44,15 @@ class Rule(ABC):
     THRESHOLDS = None
     """Thresholds that the Rule must meet to be good. See `ruleskit.thresholds.Thresholds` for more details."""
 
+    condition_index = ["features_names", "features_indexes", "bmins", "bmaxs"]
+    rule_index = ["prediction"]
+    index = condition_index + rule_index
+
+    attributes_from_test_set = []
+    attributes_from_train_set = []
+
+    daughters = []
+
     @classmethod
     def SET_THRESHOLDS(cls, path: Union[str, Path, "TransparentPath", None], show=False):
         """Set thresholds globally for all futur Rules"""
@@ -52,7 +62,9 @@ class Rule(ABC):
             cls.THRESHOLDS = Thresholds(path, show)
 
     def __init__(
-        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
+        self,
+        condition: Optional[Condition] = None,
+        activation: Optional[Activation] = None,
     ):
 
         if condition is not None and not isinstance(condition, Condition):
@@ -70,10 +82,16 @@ class Rule(ABC):
 
         self._coverage = None
         self._prediction = None
+        self._criterion = None
 
         self._time_fit = -1
+        self._time_eval = -1
         self._time_calc_activation = -1
         self._time_predict = -1
+        self._time_calc_criterion = -1
+        self._time_calc_prediction = -1
+        self._fitted = False
+        self._evaluated = False
         if self._activation is not None:
             self.check_thresholds("coverage")
 
@@ -127,7 +145,7 @@ class Rule(ABC):
         self._coverage = value
 
     def __and__(self, other: "Rule") -> "Rule":
-        """Logical AND (&) of two rules. It is simply the logical AND of the two rule's conditions and activations. """
+        """Logical AND (&) of two rules. It is simply the logical AND of the two rule's conditions and activations."""
         condition = self._condition & other._condition
         activation = self._activation & other._activation
         return self.__class__(condition, activation)
@@ -238,7 +256,9 @@ class Rule(ABC):
             return self._condition == other._condition
 
     def __contains__(self, other: "Rule") -> bool:
-        """A Rule contains another Rule if the second rule's activated points are also all activated by the first rule.
+        """
+        A Rule contains another Rule if the second rule's activated points are also all activated by the first
+        rule.
         """
         if not self._activation or not other._activation:
             return False
@@ -263,7 +283,7 @@ class Rule(ABC):
         """A Rule's length is the number of features it talks about"""
         return len(self._condition)
 
-    def evaluate(self, xs: Union["pd.DataFrame", np.ndarray]) -> Activation:
+    def evaluate_activation(self, xs: Union["pd.DataFrame", np.ndarray]) -> Activation:
         """Computes and returns the activation vector from an array of features.
 
         Parameters
@@ -281,50 +301,100 @@ class Rule(ABC):
         a = Activation(arr, to_file=self.__class__.LOCAL_ACTIVATION)
         return a
 
-    # noinspection PyUnusedLocal
-    def fit(self, xs: Union["pd.DataFrame", np.ndarray] = None, y: np.ndarray = None, **kwargs):
-        """Computes activation, and other criteria dependant on the nature of the daughter class of the Rule,
-        for a given xs and y.
+    def fit(
+        self,
+        y: Union[np.ndarray, "pd.Series"],
+        xs: Optional[Union["pd.DataFrame", np.ndarray]] = None,
+        **kwargs
+    ):
+        """Computes activation and attributes relevant to the train set
 
         Parameters
         ----------
-        xs: Union[pd:DataFrame, np.ndarray]
-            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd:DataFrame.
-        y: np.ndarray
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray.
+        y: Union[np.ndarray, pd.Series]
+        xs: Union[pd.DataFrame, np.ndarray]
         kwargs: dict
-            Other arguments used by daughter class
+            Additionnal keyword arguments for calc_<any_attribute>
         """
+        if self._fitted and xs is None:
+            return
         t0 = time()
-        if xs is not None:
-            self.calc_activation(xs)
 
-        if self._activation is None:
-            raise ValueError("If fitting without specifying xs, activation must have been computed already.")
-        self.calc_attributes(y, **kwargs)
-        if self.prediction is None:
-            raise ValueError("'fit' did not set 'prediction' : did you overload 'calc_attributes' correctly ?")
+        def launch_method(method, **kw):
+            expected_args = list(inspect.signature(method).parameters)
+            kw = {item: kw[item] for item in kw if item in expected_args}
+            method(**kw)
 
+        self.calc_activation(xs=xs)
+
+        for attr in self.__class__.attributes_from_train_set:
+            if attr == "activation":
+                raise ValueError("'activation' can not be specified in 'attributes_from_train_set'")
+            launch_method(getattr(self, f"calc_{attr}"), y=y, xs=xs, **kwargs)
+            self.check_thresholds(attr)
+            if not self.good:
+                self._time_fit = time() - t0
+                self._fitted = True
+                return
+        self.check_thresholds()
+        self.trigger_subattributes_computation()
         self._time_fit = time() - t0
+        self._fitted = True
 
-    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
-        """Implement in daughter class. Must set self._prediction."""
-        raise NotImplementedError("To implement in daughter class")
-
-    def calc_activation(self, xs: Union["pd.DataFrame", np.ndarray]):
-        """Uses self.evaluate to set self._activation.
+    def eval(
+        self,
+        y: Union[np.ndarray, "pd.Series"],
+        xs: Optional[Union["pd.DataFrame", np.ndarray]] = None,
+        recompute_activation: bool = False,
+        **kwargs,
+    ):
+        """Computes prediction, standard deviation, and regression criterion
 
         Parameters
         ----------
-        xs: Union[pd:DataFrame, np.ndarray]
-            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
-            or pd:DataFrame.
+        y: Union[np.ndarray, pd.Series]
+        xs: Union[pd.DataFrame, np.ndarray]
+        recompute_activation: bool
+            To reset self.activation using the given xs
+        kwargs
+            Additionnal keyword arguments for calc_<any_attribute>
         """
         t0 = time()
-        self._activation = self.evaluate(xs)
-        self._time_calc_activation = time() - t0
-        self.check_thresholds("coverage")
+
+        def launch_method(method, **kw):
+            expected_args = list(inspect.signature(method).parameters)
+            kw = {item: kw[item] for item in kw if item in expected_args}
+            method(**kw)
+
+        if not self.activation_available:
+            raise ValueError("Must have fitted the rule before calling 'eval'")
+
+        if recompute_activation:
+            self.calc_activation(xs=xs)
+            xs = None
+        if xs is not None:
+            activation = self.evaluate_activation(xs).raw
+        else:
+            activation = self.activation
+
+        for attr in self.__class__.attributes_from_test_set:
+            if attr == "activation":
+                raise ValueError("'activation' can not be specified in 'attributes_from_test_set'")
+            launch_method(getattr(self, f"calc_{attr}"), y=y, xs=xs, activation=activation, **kwargs)
+            self.check_thresholds(attr)
+            if not self.good:
+                self._time_fit = time() - t0
+                self._fitted = True
+                return
+        self.check_thresholds()
+        self._time_eval = time() - t0
+        self._evaluated = True
+
+    def trigger_subattributes_computation(self):
+        """Uses getattr(self, attr) to trigger important attributes computation. Important attributes should be
+        Ruleset.rule_index"""
+        for attr in self.__class__.index:
+            _ = getattr(self, attr)
 
     def predict(self, xs: Optional[Union["pd.DataFrame", np.ndarray]] = None) -> Union[np.ndarray, "pd.Series"]:
         """Returns the prediction vector. If xs is not given, will use existing activation vector.
@@ -344,24 +414,26 @@ class Rule(ABC):
         """
         t0 = time()
         if xs is not None:
-            self.calc_activation(xs)
+            act = self.evaluate_activation(xs).raw
         elif self.activation is None:
             raise ValueError("If the activation vector has not been computed yet, xs can not be None.")
-        act = self.activation
+        else:
+            act = self.activation
         to_ret = np.array([np.nan] * len(act))
         if isinstance(self.prediction, str):
             if self.prediction == "nan":
-                raise ValueError("Prediction should not be the 'nan' string, it will conflict with NaNs."
-                                 "Rename your class.")
+                raise ValueError(
+                    "Prediction should not be the 'nan' string, it will conflict with NaNs. Rename your class."
+                )
             to_ret = to_ret.astype(str)
         to_ret[act == 1] = self.prediction
-        self._time_predict = time() - t0
         if xs is not None and not isinstance(xs, np.ndarray):
-            return xs.__class__(index=xs.index, data=to_ret).squeeze()  # So not to requier pandas explicitly
+            to_ret = xs.__class__(index=xs.index, data=to_ret).squeeze()  # So not to requier pandas explicitly
+        self._time_predict = time() - t0
         return to_ret
 
     def get_correlation(self, other: "Rule") -> float:
-        """ Computes the correlation between self and other
+        """Computes the correlation between self and other
         Correlation is the number of points in common between the two vectors divided by their length, times the product
         of the rules' signs.
         Both vectors must have the same length.
@@ -372,22 +444,46 @@ class Rule(ABC):
         sign = (self.prediction / abs(self.prediction)) * (other.prediction / abs(other.prediction))
         return self._activation.get_correlation(other._activation) * sign
 
+    def calc_activation(self, xs: Union["pd.DataFrame", np.ndarray, None] = None):
+        """Uses self.evaluate to set self._activation.
+
+        Parameters
+        ----------
+        xs: Union["pd.DataFrame", np.ndarray, None]
+            The features on which the check whether the rule is activated or not. Must be a 2-D np.ndarray
+            or pd:DataFrame.
+        """
+        if xs is None:
+            if self._activation is None:
+                raise ValueError(
+                    "If calling calc_activation without specifying xs, activation must have been computed already."
+                )
+            return
+        t0 = time()
+        self._activation = self.evaluate_activation(xs)
+        self._time_calc_activation = time() - t0
+        self.check_thresholds("coverage")
+
 
 # noinspection PyUnresolvedReferences
 class RegressionRule(Rule):
-
     """Rule applied on continuous target data."""
 
+    rule_index = Rule.rule_index + ["coverage", "criterion", "std"]
+    index = Rule.condition_index + rule_index
+    attributes_from_test_set = ["criterion"]
+    attributes_from_train_set = Rule.attributes_from_train_set + ["prediction", "std", "sign"]
+
     def __init__(
-        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
+        self,
+        condition: Optional[Condition] = None,
+        activation: Optional[Activation] = None,
     ):
         super().__init__(condition, activation)
         self._std = None
-        self._criterion = None
+        self._sign = None
 
         # Inspection / Audit attributs
-        self._time_calc_criterion = -1
-        self._time_calc_prediction = -1
         self._time_calc_std = -1
 
     @property
@@ -410,86 +506,86 @@ class RegressionRule(Rule):
     def time_calc_std(self):
         return self._time_calc_std
 
-    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
-        """Computes prediction, standard deviation, and regression criterion
-        
-        Parameters
-        ----------
-        y: Union[np.ndarray, pd.Series]
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
-            or pd.Series.
-        kwargs: dict
-            Arguments for calc_regression_criterion
-        """
-        self.calc_prediction(y)
-        self.calc_std(y)
-        prediction_vector = self.prediction * np.where(self.activation == 0, np.nan, self.activation)
-        self.calc_criterion(prediction_vector, y, **kwargs)
-
-    def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
+    def calc_prediction(self, y: [np.ndarray, "pd.Series"], activation: Optional[np.ndarray] = None):
         """Computes the mean of all activated points in target y and use it as prediction
-        
+
         Parameters
         ----------
         y: [np.ndarray, pd.Series]
             The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
             or pd.Series
+        activation: Optional[np.ndarray]
+            If specified, uses this activation instead of self.activation
         """
         t0 = time()
-        if self.activation is None:
+        if activation is None:
+            activation = self.activation
+        if activation is None:
             return None
-        self._prediction = functions.conditional_mean(self.activation, y)
+        self._prediction = functions.conditional_mean(activation, y)
         self._time_calc_prediction = time() - t0
         self.check_thresholds("prediction")
 
-    def calc_std(self, y: Union[np.ndarray, "pd.Series"]):
+    def calc_sign(self):
+        if self._prediction is None:
+            return
+        if self._prediction < 0:
+            self._sign = "-"
+        else:
+            self._sign = "+"
+
+    def calc_std(self, y: Union[np.ndarray, "pd.Series"], activation: Optional[np.ndarray] = None):
         """Computes the standard deviation of all activated points in target y
-        
+
         Parameters
         ----------
         y: Union[np.ndarray, pd.Series]
             The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
             or pd.Series.
+        activation: Optional[np.ndarray]
+            If specified, uses this activation instead of self.activation
         """
         t0 = time()
-        if self.activation is None:
+        if activation is None:
+            activation = self.activation
+        if activation is None:
             return None
-        self._std = functions.conditional_std(self.activation, y)
+        self._std = functions.conditional_std(activation, y)
         self._time_calc_std = time() - t0
         self.check_thresholds("std")
 
-    def calc_criterion(self, p: Union[np.ndarray, "pd.Series"], y: Union[np.ndarray, "pd.Series"], **kwargs):
+    def calc_criterion(self, y: Union[np.ndarray, "pd.Series"], activation: Optional[np.ndarray] = None, **kwargs):
         """
         Parameters
         ----------
-        p: Union[np.ndarray, pd.Series]
-            Prediction vector. Must be a 1-D np.ndarray or pd.Series.
         y: Union[np.ndarray, pd.Series]
             The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
             or pd.Series.
+        activation: Optional[np.ndarray]
+            If specified, uses this activation instead of self.activation
         kwargs: dict
             Arguments for calc_regression_criterion
         """
         t0 = time()
-        self._criterion = functions.calc_regression_criterion(p, y, **kwargs)
+        if activation is None:
+            activation = self.activation
+        if activation is None:
+            return None
+        self._criterion = functions.calc_regression_criterion(
+            self.prediction * np.where(activation == 0, np.nan, activation), y, **kwargs
+        )
         self._time_calc_criterion = time() - t0
         self.check_thresholds("criterion")
 
 
 # noinspection PyUnresolvedReferences
 class ClassificationRule(Rule):
-
     """Rule applied on discret target data."""
 
-    def __init__(
-        self, condition: Optional[Condition] = None, activation: Optional[Activation] = None,
-    ):
-        super().__init__(condition, activation)
-
-        self._criterion = None
-
-        self._time_calc_criterion = -1
-        self._time_calc_prediction = -1
+    rule_index = Rule.rule_index + ["coverage", "criterion"]
+    index = Rule.condition_index + rule_index
+    attributes_from_test_set = ["criterion"]
+    attributes_from_train_set = Rule.attributes_from_train_set + ["prediction"]
 
     @property
     def prediction(self) -> Union[int, str, np.integer, np.float, None]:
@@ -509,19 +605,6 @@ class ClassificationRule(Rule):
     def criterion(self) -> float:
         return self._criterion
 
-    def calc_attributes(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
-        """
-        Parameters
-        ----------
-        y: Union[np.ndarray, pd.Series]
-            The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
-            or pd.Series.
-        kwargs: dict
-            Arguments for calc_classification_criterion
-        """
-        self.calc_prediction(y)
-        self.calc_criterion(y, **kwargs)
-
     def calc_prediction(self, y: [np.ndarray, "pd.Series"]):
         """
         Parameters
@@ -537,17 +620,23 @@ class ClassificationRule(Rule):
         self._time_calc_prediction = time() - t0
         self.check_thresholds("prediction")
 
-    def calc_criterion(self, y: Union[np.ndarray, "pd.Series"], **kwargs):
+    def calc_criterion(self, y: Union[np.ndarray, "pd.Series"], activation: Optional[np.ndarray] = None, **kwargs):
         """
         Parameters
         ----------
         y: Union[np.ndarray, pd.Series]
             The targets on which to evaluate the rule prediction, and possibly other criteria. Must be a 1-D np.ndarray
             or pd.Series
+        activation: Optional[np.ndarray]
+            If specified, uses this activation instead of self.activation
         kwargs: dict
             Arguments for calc_classification_criterion
         """
         t0 = time()
-        self._criterion = functions.calc_classification_criterion(self.activation, self.prediction, y, **kwargs)
+        if activation is None:
+            activation = self.activation
+        if activation is None:
+            return None
+        self._criterion = functions.calc_classification_criterion(activation, self.prediction, y, **kwargs)
         self._time_calc_criterion = time() - t0
         self.check_thresholds("criterion")
